#ifndef TSYNC_OBJECT_H
#define TSYNC_OBJECT_H

/////////////////////////////////////////////////////////////////////////////
//
// File: TsynObject.h
// Author: Kelvin Sung
// Date: April 2003
//
// 		TsyncObject: A class demonstrating some of the functionality of
// 		the MFC synchronization objects (CMutex and CEvent).
//
// **Note** = 
//
//			Mutex Locks are per-Thread property, once a thread obtains a lock
//			only that thread can "unlock", and as long as a thread has the lock
//			to a mutex, no other threads should try to unlock (this is a run time error)
//			
//
//			Event Locks are waiting for signals, (SetEvent, or PulseEvent).
//			No one thread really owns the Lock, all threads waiting on an event
//			Lock has the potential being wake up when any thread Set/PulseEvent
//			Whether all or one waiting thread gets wake up by an event depends
//			on if the CEvent is mutual or automatic reset (refer to MSDN on-line
//			help pages for more about this).
//
//	When Event is present (Think Green light) - do not block on Lock (think do not stop at intersection)
//	When NoEvent (Re-set) (Think Red Light)   - block on Lock (think stop at intersection)
//
//
// ***CMutex***
// ============
//		Changing state/inquiry:
//
//			Lock() - if its not already locked, 
//							you "get" the lock and no-one else can get it
//							until you "unlock"
//				     if its already locked,
//							you get blocked, until the thread the locks does a "unlock"
//
//			unlock() - you can only unlock a thread you have locked.
//					   this will wake up _ALL_ waiting threads (those who tried to lock)
//					   and only one thread will get the next lock
//			
//
// ***CEvent***
// ============
//
//		Changing state:
//
//			SetEvent() - Event present (Green Light), all blocked threads can go now
//			ResetEvent() - Remove Event (Red Light), all future Lock() gets blocked
//
//
//		Inquiring about state:
//
//			Lock() - check for the presence of event (at intersection)
//					 if event is present (someone did a SetEvent()), thread can continue
//					 if event is not present, this will block the thread
//
////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////
//
//			fCommand - is a number (set by main thread)
//			fCommandLock - is a CMutex allowing exclusive 
//						   access to the fCommand
//
//
//	
//
// main -> SetCommand()
//				this makes sure previous command is cleared (fCommandReadEvent), or else this blocks.
//					=> Send Raise NextCommandEvent to all waiting workers
//
//
// All Workers -> CheckNextCommand()
//					if not mine, wait on fCommandRead event and then go check on command again.
//					Note: before CommandRead event comes, the id-ed worker has not read the
//					command yet, meaning if we go back, we will see a valid command that is
//					not ours. A waste of processing time, so we wait for current valid commnad
//					that we have checked and know is not ours to be read by the appropriate
//					worker. 
//
//
// The id-ed worker -> ClearCommand and Send CommandRead event to wake up all waiting workers and main
//
//

#include <iostream.h>

#include <afxmt.h>
		// for MFC Synchronization objects (CEvent/CMutex etc.)

#define kAllThreadQuits -10

class TsyncObject {
	private:

		int			fCommand;
						//  main thread sets this command to a thread id,
						//  the thread that is identified by this id should
						//  clear this command (to -1) and then go to work.
						//
		CMutex  	*fCommandMutex;
						//  make sure no two threads will write on the 
						//	fCommand at the same time.

		CEvent		*fCommandReadEvent;
						// prevents mainThread from writing into fCommand before
						// worker threads have chance to see the fCommand.
						// after worker thread sees the fCommand, this Event
						// will be set for main thread to see that we are
						// ready for next command
						//

		CEvent  	*fNextCommandEvent;
						// Threads will all synchronized on this event, waiting
						// for next command, when next command comes, all 
						// threads will be wake up, they will check for fCommand
						// the identify thread will do the necessary work, all
						// other threads will go right back to waiting
						//
						// To support the above functionality, we must set
						// this CEvent object to:Manual Reset 
						// so that we can ensure _ALL_ waiting
						// threads will be wake up to check for commands.
						//
						// Please refer to the hand-outs, with Automatic reset
						// and PulseEvent, only one thread will proceed at any 
						// time.
						//

		void ChangeCommand(int id)
			// Change fCommand we make sure we check
			// for exclusive right.
			{
				#ifdef _DEBUG_DETAIL
						cerr << "... ChangeCommand id(" << id << "): Trying to get lock" << endl;
				#endif
				
				// Try to gain exclusive access to fCommand (worker may be
				// writting to this right now)
				// this may block!!
				fCommandMutex->Lock();
			

				// At this point we
				// have exclusive access to the fCommand
				fCommand = id;
					// change the fCommand
					//

				#ifdef _DEBUG_DETAIL
						cerr << "... ChangeCommand changed, releasing lock... " << endl;
				#endif

				fCommandMutex->Unlock();
			}
				
	public:
		
		TsyncObject() :
				fCommand(-1)
			{
				//
				// fCommand Mutex/Lock set ...
				fCommandMutex = new CMutex(); 				
				
				//
				// fCommandRead Event/Lock set
				fCommandReadEvent = new CEvent(true,	// initial condition is true (event present - previous command has been read)
												true); 	// Must be manual reset
				//
				// fNextCommand Event/Lock set
				fNextCommandEvent = new CEvent(false,	 // initial condition is false (no event present - no command available for workers)
												true);	 // we want manual reset (to allow all threads wake up)

				fCommandMutex->Unlock();				// initially command is _NOT_ locked
			}


		void SetCommand(int id)
			// Change fCommand to the id, and wakes up all
			// waiting threads .
			// If cannot get exclusve right to use fCommand
			// this may block!
			{
				
				#ifdef _DEBUG_DETAIL
						cerr << "... SetCommand Getting Lock... " << endl;
				#endif

				// 1. Make sure command is already read by the worker Thread.
				// if we have previous set the fCommand and the appropriated
				// worker has not read the fCommand, this will block!!
				//				
				fCommandReadEvent->Lock();

				//////////////////////////////////////////////////////////
				//
				// Now, we can set new commands for the worker threads
				//
				//////////////////////////////////////////////////////////

				// Try to change fCommand, if we do not have exclusive right
				// this function call may block!!
				ChangeCommand(id);
		

				fCommandReadEvent->ResetEvent();
					// A new command has not been read yet.

				fNextCommandEvent->SetEvent();
					// Wake up all waiting threads to check
			}

		

		void ClearCommand() 
			// Appropriate worker thread will call this function
			// to clear fCommand. 
			// 
			// **** Notice if ClearCommand() is not
			// called, then TsynObject basically locks up and no
			// more communication can happen! **************
			{
				
				fNextCommandEvent->ResetEvent();
					// we have seen the next command this event is responded to it

				#ifdef _DEBUG_DETAIL
						cerr << "... Clearing fCommand ... " << endl;
				#endif

				ChangeCommand(-1);	
					// clears the fCommand

				#ifdef _DEBUG_DETAIL
						cerr << "... Cleared fCommand to -1... " << endl;
				#endif

				fCommandReadEvent->SetEvent();	
					// tells other workers and main threads that fCommand has been read
			}


		bool CheckNextCommand(int id) 
			//
			// wait till next command is there
			// when the command is there, check if
			// it is for us, if it is, clear the state for
			// main to issue next command else, return false
			//
			{
				bool found = false;

				
				#ifdef _DEBUG_DETAIL
						cerr << "Thread(" << id << "): Waiting ... " << endl;
				#endif

				fNextCommandEvent->Lock();
					// Check for next Command, will block if next command
					// is not ready


				if (id == fCommand) {
					#ifdef _DEBUG_DETAIL
						cerr << "Thread(" << id << "): Match!" << endl;
					#endif
					found = true;
					ClearCommand();
					
					#ifdef _DEBUG_DETAIL
						cerr << "	Cleared fCommand ready to return" << endl;
					#endif

				} else {

					//
					// Current command is not ours, lets wait for this one
					// being read.
					//
					fCommandReadEvent->Lock();
				}

				return found;
			}


		void SetTerminateAll() { ChangeCommand(kAllThreadQuits); }
		bool ShouldQuit() { return fCommand == kAllThreadQuits; }

};

#endif
